#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>
#include <netdb.h>

#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>

#include <netinet/in.h>
#include <arpa/inet.h>

#if 1
#define DP(X)	do { } while (0)
#else
#include <stdarg.h>
#define DP(X)	do { dpprintf X; } while (0)
void
dpprintf(const char *format, ...)
{
  va_list	list;

  fprintf(stderr, "[");
  va_start(list, format);
  vfprintf(stderr, format, list);
  va_end(list);
  fprintf(stderr, "]\n");
  fflush(stderr);
}
#endif

static void
ex(const char *s)
{
  perror(s);
  fflush(stderr);
  exit(2);
}

static void
getaddr(struct sockaddr_in *sin, const char *host, int port)
{
  memset(sin, 0, sizeof *sin);
  sin->sin_family	= AF_INET;
  sin->sin_addr.s_addr	= htonl(INADDR_ANY);
  sin->sin_port		= htons(port);
  if (host && !inet_aton(host, &sin->sin_addr))
    {
      struct hostent	*he;

      if ((he=gethostbyname(host))==0)
	ex(host);
      if (he->h_addrtype!=AF_INET || he->h_length!=sizeof sin->sin_addr)
	ex("unsupported host address");
      memcpy(&sin->sin_addr, he->h_addr, sizeof sin->sin_addr);
    }
}

static int
sockopt(int sock)
{
  int			on;
  struct linger		linger;

  on = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof on))
    ex("setsockopt reuse");
  linger.l_onoff	= 1;
  linger.l_linger	= 65535;
  if (setsockopt(sock, SOL_SOCKET, SO_LINGER, &linger, sizeof linger))
    ex("setsockopt linger");
  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &on, sizeof on))
    ex("setsockopt sndbuf");
  on	= 102400;
  if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &on, sizeof on))
    ex("setsockopt rcvbuf");

#if 0
  if (fcntl(sock, F_SETFL, O_NDELAY))
    ex("O_NDELAY");
#endif

  return sock;
}

static int
sockopen(const char *host, int port)
{
  int			sock;
  struct sockaddr_in	sin;

  sock	= socket( AF_INET, SOCK_STREAM, 0 );
  if (sock<0)
    ex("socket");
  getaddr(&sin, host, port);

  if (bind(sock, (struct sockaddr *)&sin, sizeof sin))
    ex("bind");

  return sockopt(sock);
}

static int
do_connect(const char *host, int port)
{
  struct sockaddr_in	sin;
  int			sock;

  sock	= sockopen(NULL, 0);

  getaddr(&sin, host, port);
  if (connect(sock, &sin, sizeof sin))
    ex("connect");
  return sock;
}

static void
closew(int fd)
{
  if (fd>=0)
    {
      shutdown(fd, 1);
      close(fd);
    }
}

static void
closer(int fd)
{
  if (fd>=0)
    {
      shutdown(fd, 0);
      close(fd);
    }
}

static void
selectloop(int in, int sock, int out,
	   int throttle, int mtu)
{
  struct transfer
  {
    int		ok;
    int		from, to;
    int		pos, fill;
    long	throttle;
    char	buf[BUFSIZ];
  } move[2];
  int			max, block, sock2;
  struct timeval	now;
  long			throtmul, throtdiv, throtmax;

  if (throttle && gettimeofday(&now, NULL))
    {
      perror("gettimeofday");
      throttle	= 0;
    }

  sock2			= dup(sock);
  if (sock2<0)
    ex("dup");
  sockopt(sock2);

  block	= BUFSIZ;
  if (throttle && block>throttle)
    block	= throttle;
  if (!mtu || mtu>BUFSIZ)
    mtu	= BUFSIZ;
  if (throttle && mtu>throttle)
    mtu	= throttle;

  throtmax	= throttle/5;
  if (throtmax<mtu)
    throtmax	= mtu;
  if (throttle && throtmax>throttle)
    throtmax	= throttle;

  throtdiv	= 1000;
  throtmul	= throttle;
  while (throtdiv>1 && throtmul>=1000000)
    {
      throtmul	/= 10;
      throtdiv	/= 10;
    }

  move[0].ok		= 1;
  move[0].from		= in;
  move[0].to		= sock;
  move[0].pos		= 0;
  move[0].fill		= 0;
  move[0].throttle	= throtmax;

  move[1].ok		= 1;
  move[1].from		= sock2;
  move[1].to		= out;
  move[1].pos		= 0;
  move[1].fill		= 0;
  move[1].throttle	= throtmax;

  max	= sock2;
  if (max<=sock)
    max	= sock;
  if (max<=in)
    max	= in;
  if (max<=out)
    max	= out;
  max++;

  DP(("throtmul=%ld throtdiv=%ld block=%d mtu=%d max=%d",
      throtmul, throtdiv, block, mtu, max));

  while (move[0].ok || move[1].ok)
    {
      fd_set		fdin, fdout, fdex;
      struct timeval	timeout, *timeoutp, *timeoutn;
      int		i;
      long		throtval;

      FD_ZERO(&fdin);
      FD_ZERO(&fdout);
      FD_ZERO(&fdex);

      timeoutp	= 0;
      timeoutn	= 0;
      throtval	= 0;
      if (throttle)
	{
	  struct timeval	tmp;
	  long			delta;

	  gettimeofday(&tmp, NULL);
	  delta	= (tmp.tv_usec-now.tv_usec)/1000;
	  if (tmp.tv_sec!=now.tv_sec)
	    {
	      now.tv_sec	= tmp.tv_sec-1;
	      delta		+= 1000;
	    }

	  throtval	= delta*throtmul/throtdiv;
	  if (throtval<0)
	    throtval	= 0;
	  if (throtval>=throtmax)
	    throtval	= throtmax;
	  DP(("throtval %ld", throtval));

	  delta		= throtval*throtdiv/throtmul;
	  DP(("delta %ld", delta));
	  now.tv_usec	+= delta*1000;
	  if (now.tv_usec>=1000000)
	    {
	      now.tv_sec	+= 1;
	      now.tv_usec	-= 1000000;
	    }

	  timeout.tv_sec	= 0;
	  timeout.tv_usec	= 100000;
	  timeoutn		= &timeout;
	}

      for (i=2; --i>=0; )
	if (move[i].ok)
	  {
	    if (move[i].throttle<throttle)
	      {
		move[i].throttle	+= throtval;
		DP(("throttle %d=%d", i, move[i].throttle));
		timeoutp		=  timeoutn;
	      }
	    if (move[i].fill>move[i].pos)
	      {
		if (move[i].throttle>0)
		  {
		    DP(("set out %d", move[i].to));
		    FD_SET(move[i].to, &fdout);
		  }
	      }
	    else if (move[i].from<0)
	      {
		DP(("finish %d", i));
		closew(move[i].to);
		move[i].ok	= 0;
		continue;
	      }
	    FD_SET(move[i].to, &fdex);
	    if (move[i].from>=0)
	      {
		FD_SET(move[i].from, &fdex);
		if (move[i].pos || move[i].fill<block)
		  {
		    DP(("set in %d", move[i].from));
		    FD_SET(move[i].from, &fdin);
		  }
	      }
	  }

      if (!timeoutp && !move[0].ok && !move[1].ok)
	break;

      if (select(max, &fdin, &fdout, &fdex, timeoutp)<0)
	{
	  if (errno==EINTR)
	    continue;
	  ex("select");
	}

      for (i=2; --i>=0; )
	if (move[i].ok)
	  {
	    if (FD_ISSET(move[i].to, &fdout))
	      {
		int	put;

		DP(("put %d", move[i].to));
		put	= move[i].fill-move[i].pos;
		if (put>mtu)
		  put	= mtu;
		if (put>move[i].throttle)
		  put	= move[i].throttle;
		DP(("toput %d", put));
		put	= write(move[i].to, move[i].buf+move[i].pos, put);
		DP(("putted %d", put));
		if (put>=0)
		  {
		    move[i].pos	+= put;
		    if (throttle)
		      move[i].throttle	-= put;
		  }
		else if (errno!=EINTR)
		  {
		    DP(("close %d", i));
		    closew(move[i].to);
		    closer(move[i].from);
		    move[i].ok	= 0;
		    break;
		  }
	      }
	    else if (FD_ISSET(move[i].to, &fdex))
	      {
		DP(("ex %d", move[i].to));
		closew(move[i].to);
		closer(move[i].from);
		move[i].ok	= 0;
		break;
	      }
	    if (move[i].from>=0)
	      if (FD_ISSET(move[i].from, &fdin))
		{
		  int	get;

		  DP(("get %d", move[i].from));
		  if (move[i].fill<=move[i].pos)
		    {
		      move[i].pos		= 0;
		      move[i].fill	= 0;
		    }
		  if (move[i].pos && move[i].fill+mtu>BUFSIZ)
		    {
		      int	len;

		      len		= move[i].fill-move[i].pos;
		      memmove(move[i].buf, move[i].buf+move[i].pos, len);
		      move[i].pos	= 0;
		      move[i].fill	= len;
		    }
		  get	= BUFSIZ-move[i].fill;
		  DP(("toget %d", get));
		  if (get>0)
		    {
		      get	= read(move[i].from, move[i].buf+move[i].fill, get);
		      DP(("got %d", get));
		      if (get>0)
			move[i].fill	+= get;
		      else if (!get || errno!=EINTR)
			{
			  DP(("shutdown %d", i));
			  closer(move[i].from);
			  move[i].from	= -1;
			  break;
			}
		    }
		}
	      else if (FD_ISSET(move[i].from, &fdex))
		{
		  DP(("ex %d", move[i].from));
		  closer(move[i].from);
		  move[i].from	= -1;
		  break;
		}
	  }
    }
}

int
main(int argc, char **argv)
{
  int	mtu, throttle;

  mtu		= BUFSIZ;
  throttle	= 0;
  switch (argc)
    {
    case 5:
      mtu	= atoi(argv[4]);
    case 4:
      throttle	= atoi(argv[3]);
    case 3:
      break;

    default:
      fprintf(stderr, "Usage: %s host port [Byte/s [MTU]]\n", argv[0]);
      return 1;
    }

#if 1
  freopen("/tmp/portforward.debug", "a", stderr);
#endif

  sockopt(0);
  sockopt(1);
  selectloop(0, do_connect(argv[1], atoi(argv[2])), 1, throttle, mtu);
  return 0;
}
