#include <stdio.h>
#include <time.h>
#include <stdarg.h>
#include <errno.h>

#include <signal.h>

#include "tino.h"
#include "tinostr.h"
#include "tinosock.h"

#define PROGRAM_NAME	"cfp"
#define PROGRAM_VERSION	"v-0.9"

static struct str	outp;

static char *
lasts(char *s)
{
  size_t	len;

  DP(("lasts(%s)", s));
  len	= strlen(s);
  while (len)
    if (s[--len]==' ')
      {
	char	*ret;

	DP(("lasts() %d %s", len, s+len));
	ret	= s+len+1;
	do
	  s[len]	= 0;
	while (len && s[--len]==' ');
	DP(("lasts() %d %s", len, ret));
	return ret;
      }
  DP(("lasts() 0"));
  return 0;
}

static char *
nexts(char *s)
{
  size_t	len, i;

  DP(("nexts(%s)", s));
  len	= strlen(s);
  for (i=0; i<len; i++)
    if (s[i]==' ')
      {
	DP(("nexts() %d %s", i, s+i));
	do
	  s[i]	= 0;
	while (s[++i]==' ' && i<len);
	DP(("nexts() %d %s", i, s+i));
	return s+i;
      }
  DP(("nexts() NULL"));
  return 0;
}

static void
out_start(void)
{
  str_reset(&outp);
}

static void
out(const char *s)
{
  DP(("out(%s)", s));
  str_append_n(str_append(str_def(&outp), s), "\r\n", 2);
}

static void
outfv(const char *s, va_list list)
{
  static struct str	p;

  DP(("outfv(%s,%p)", s, list));
  str_fv(str_def(&p), s, list);
  out(p.s);
  str_free(&p);
}

static void
outf(const char *s, ...)
{
  va_list	list;

  DP(("outf(%s,..)", s));
  va_start(list, s);
  outfv(s, list);
  va_end(list);
}

static struct str *
htmlescape(struct str *buf, struct str *p)
{
  char	*s;
  int	n;

  DP(("htmlescape(%p,%p)", buf, p));

  str_reset(buf);
  s	= p->s;
  n	= 0;
  for (;;)
    {
      char	*w;

      switch (s[n])
	{
	default:
	  n++;
	  continue;

	case 0:
	  w	= 0;
	  break;

	case '&':
	  w	= "&amp;";
	  break;

	case '<':
	  w	= "&lt;";
	  break;

	case '>':
	  w	= "&gt;";
	  break;
	}
      if (n)
	str_append_n(buf, s, n);
      if (!w)
	break;
      str_append(buf, w);
      s	+= n+1;
      n	= 0;
    }
  return buf;
}

static void
err(int status, const char *title, char *param, ...)
{
  time_t	t;
  char		buf[256];
  va_list	list;

  if (!status)
    status	= 400;
  if (!title)
    title	= "general proxy error";

  time(&t);
  strftime(buf, sizeof buf, "%a, %d %b %Y %H:%M:%S GMT", gmtime(&t));

  printf("HTTP/1.0 %d %s\r\n", status, title);
  printf("Server: %s %s \r\n", PROGRAM_NAME, PROGRAM_VERSION);
  printf("Date: %s\r\n", buf);
  printf("Connection: close\r\n");

  out_start();
  out("<HTML>");
  out("<HEAD>");
  outf("<TITLE>%d %s</TITLE>", status, title);
  out("</HEAD>");
  out("<BODY>");
  outf("<H1>%d %s</H1>", status, title);
  if (param)
    {
      struct str *a, *b;

      va_start(list, param);
      a	= str_fv(str_new(), param, list);
      va_end(list);
      b	= htmlescape(str_new(), a);

      str_release(a);
      out(b->s);
      str_release(b);
    }
  out("</BODY>");
  out("</HTML>");

  if (param)
    {
      va_list	list;

      va_start(list, param);
      vprintf(param, list);
      va_end(list);
      printf("\r\n");
    }

  printf("Content-Length: %ld\r\n", (long)strlen(outp.s));
  printf("\r\n");
  printf("%s", outp.s);
  fflush(stdout);
  exit(1);
}

static void
ex(const char *s)
{
  err(0, s, NULL);
}

static void
trimcr(struct str *p)
{
  size_t	len;

  len	= strlen(p->s);
  if (len && p->s[len-1]=='\r')
    p->s[len-1]	= 0;
}

/* XXX add header filter etc. here XXX
 */
static long
copyhead(FILE *out, FILE *in)
{
  struct str	*p;
  long		content;

  DP(("copyhead(%p,%p)"));
  content	= -1;
  p		= str_new();
  for (;;)
    {
      str_getline(p, in);
      trimcr(p);
      if (!p->s[0])
	break;
      
      /* weed out unknown headers here
       */
      000;

      if (!strncasecmp(p->s, "content-length:", 15))
	content	= strtol(p->s+15, NULL, 10);

      fprintf(out, "%s\r\n", p->s);
    }
  str_release(p);

  return content;
}

/* XXX add data output filter etc. here XXX
 */
static void
copybody(FILE *out, FILE *in, long n)
{
  char	buf[BUFSIZ];
  int	oops;

  DP(("copybody(%p,%p,%ld)", out, in, n));
  oops	= 0;
  while (n && !feof(in) && !ferror(in))
    {
      int	got, max;

      max	= sizeof buf/sizeof *buf;
      if (max>n)
	max	= (int)n;
      got	= fread(buf, 1, max, in);
      if (got<=0)
	{
	  if (++oops>1000)
	    err(0, "internal error", "read looping");
	  continue;
	}
      if (got>max)
	err(0, "internal error", "library fread bug");

      oops	 =0;
      /* If you library is buggy, fwrite() does not always write all data!
       */
      000;
      fwrite(buf, 1, got, out);
      n	-= got;
    }
}

static void
directcopy(FILE *out, FILE *in)
{
  char	buf[BUFSIZ];
  int	oops;
  int	n;

  oops	= 0;
  while ((n=fread(buf, 1, sizeof buf, in))>=0)
    {
      if (!n)
	{
	  if ((errno!=EINTR && errno!=EAGAIN) || ++oops>1000)
	    break;
	  continue;
	}
      oops	= 0;
      if (fwrite(buf, 1, n, out)<=0)
	break;
      fflush(out);
    }
}

static const char *
shell_escape(const char *s)
{
  const char	*p;
  struct str	*w;
  int		i;

  DP(("shell_escape(%s)", s));
  if (!s)
    return "";
  if ((p=strchr(s, '\''))==0)
    return s;
  w	= str_new();
  i	= 0;
  do
    {
      str_append_n(w, s+i, p-s-i);
      i=p-s;
      str_append(w, "'\"'");
      while (*p++=='\'');
      str_append_n(w, s+i, --p-s-i);
      i=p-s;
      str_append(w, "\"'");
    }
  while ((p=strchr(s+i, '\''))!=0);
  return str_append(w, s+i)->s;
}

static int
forknonhttp(const char *meth, const char *scheme, const char *auth,
	    const char *host, int port, const char *path, pid_t *pid_p)
{
  int	so[2];
  pid_t	pid;

  DP(("forknonhttp(%s,%s,%s,%s,%d,%s,%p)", meth, scheme, auth, host, port, path, pid_p));
  /* as we are a simple proxy,
   * we have to fork a helper application to process
   * complex types (like ftp)
   */
  if (socketpair(PF_UNIX, SOCK_STREAM, 0, so))
    return 1;
  
  if ((pid=fork())==0)
    {
      struct str *f;

      close(so[0]);
      dup2(so[1], 0);
      dup2(so[1], 1);
      DP(("fork"));
      f = str_f(str_new(),
		"/usr/local/sbin/cfp-helper.sh '%s' '%s' '%s' '%s' %d '%s'",
		shell_escape(meth),
		shell_escape(scheme),
		shell_escape(auth),
		shell_escape(host),
		port,
		shell_escape(path));
      DP(("execl sh -c %s", f->s));
      execl("/bin/sh", "sh", "-c", f->s, NULL);
      DP(("execl oops"));
      err(500, "cannot execl", NULL);
      exit(1);
    }
  if (pid_p)
    *pid_p	= pid;
  close(so[1]);
  DP(("forknonhttp %d %d", so[0], pid));
  return so[0];
}

static void
doget(char *url, int body, int simple, const char *meth)
{
  char		*s, *path, *prefix, *host;
  int		port;
  struct str	*p;
  int		sock;
  FILE		*in, *out;
  long		content;
  pid_t		pid;

  DP(("doget(%s,%d,%d,%s)", url, body, simple, meth));

  /* RFC2396 3.1
   * We allow more than in the RFC yet.
   * It's then checked below.
   */
  host	= strchr(url, ':');
  if (!host || host[1]!='/' || host[2]!='/')
    err(0, "bad request", "unparsable URL %s", url);
  *host	=  0;
  host	+= 3;

  /* get the path
   * RFC2396 3.2
   * Note that / is prepended to the path unconditionally
   * RFC2616, 5.1.2
   */
  path	= strpbrk(host, "/?");
  if (!path)
    path	= "";
  else if (*path=='?')
    {
      *path++	= 0;
      /* we now lack the ?
       * and we have the special case that abs_path is empty, too (RFC2616, 5.1.2)
       * but this is alredy handled.
       */
      path	= str_append(str_append(str_new(), "?"), path)->s;
    }
  else
    *path++	= 0;
  DP(("path=%s", path));

#if 0
  /* gracefully discard blanks
   * or ..
   * XXX promote them to %20? XXX
   */
  000;
  if ((s=strchr(path, ' '))!=0)
    *s		= 0;
#endif

  /* user:pass@ part
   * RFC2396 3.2.2
   */
  prefix	= host;
  host		= strrchr(host, '@');
  if (host)
    *host++	= 0;
  else
    {
      host	= prefix;
      prefix	= 0;
    }

  /* fetch the port
   * RFC2396 3.2.2
   * XXX this currently leaves us alone with IPv6 IP numbers! XXX
   */
  port		= 0;
  if ((s=strchr(url, ':'))!=0)
    {
      *s++	= 0;
      port	= strtol(s, &s, 10);
      /* gracefully ignore garbage?
       */
      000;
    }

  000; /* Add security here */


  /* Open HTTP stream
   * If we have proprietary schemes here
   * we have to fork a HTTP server emulating software
   * to hande the scheme.
   */
  if (!strcasecmp(url, "http"))
    sock	= do_connect(host, (port ? port : 80));
  else if ((sock=forknonhttp(meth, url, prefix, host, port, path, &pid))<0)
    err(0, "bad request", "unknown scheme %s", url);


  if ((out=fdopen(sock, "w"))==NULL)
    err(0, "cannot fdopen", "w");
  if ((in=fdopen(sock, "r"))==NULL)
    err(0, "cannot fdopen", "r");

  DP(("doget() request %s /%s", meth, path));
  /* send request
   */
  fprintf(out, "%s /%s HTTP/1.0\r\n", meth, path);
  if (simple)
    {
      DP(("doget() simple"));
      fprintf(out, "\r\n");
    }
  else
    {
      DP(("doget() copyhead"));
      content	= copyhead(out, stdin);
      fprintf(out, "\r\n");
      DP(("doget() copybody"));
      if (content>0)
	copybody(out, stdin, content);
      DP(("doget() here"));
    }
  if (feof(out) || ferror(out))
    err(500, "server closed connection", NULL);

  DP(("doget() fflush(%p)", out));
  /* See signal(SIGPIPE, SIG_IGN) below ..
   */
  fflush(out);

  fDP(("doget() response"));
  /* get response
   */
  p	= str_getline(str_new(), in);
  if (!*p->s)
    {
      /* XXX retry? XXX */
      000;
      err(500, "null response", NULL);
    }

  printf("%s\r\n", p->s);
  {
    int	i;

    i	= atoi(nexts(p->s));
    /* RFC2616 4.4 */
    if (i<200 || i==204 || i==304)
      body	= 0;
  }
  str_release(p);

  /* we should weed out "connection"-head here
   */
  content	= copyhead(stdout, in);
  printf("Connection: close\r\n");
  printf("\r\n");
  fflush(stdout);

  if (!body)
    return;

  if (content>=0)
    copybody(stdout, in, content);
  else
    directcopy(stdout, in);
}

static void
fn_get(char *url, int simple, const char *meth)
{
  doget(url, 1, simple, meth);
}

static void
fn_head(char *url, int simple, const char *meth)
{
  doget(url, 0, simple, meth);
}

static void
fn_connect(char *url, int simple, const char *meth)
{
  char	*s;
  int	port;
  int	sock;

  DP(("fn_connect(%s,%d,%s)", url, simple, meth));
  
  port		= 443;
  if ((s=strchr(url, ':'))!=0)
    {
      *s++	= 0;
      port	= strtol(s, &s, 10);
      /* gracefully ignore garbage?
       */
      000;
    }
  sock	= do_connect(url, port);

  fprintf(stdout, "HTTP/1.0 200 Connected\r\n\r\n");
  fflush(stdout);

  selectloop(0, sock, 1, 0, 0);
}

struct meth
  {
    const char	*meth;
    void	(*fn)(char *url, int simple, const char *meth);
  } meth[] =
  {
    { "GET",		fn_get },
    { "POST",		fn_get },
    { "HEAD",		fn_head },
    { "CONNECT",	fn_connect },
  };

int
main(int argc, char **argv)
{
  struct str	*s;
  char		*last, *url;
  int		i, simple;

  /* We have to ignore this
   * as sometimes socketpair(AF_UNIX, ...) sends this
   * and default is to terminate the application ..
   */
  signal(SIGPIPE, SIG_IGN);

  DP(("main(%d,%p)", argc, argv));
  str_getline(s=str_new(), stdin);

  url	= nexts(s->s);
  if (!url)
    err(0, "missing URL", NULL);
    
  simple	= 0;
  if ((last=lasts(url))==NULL || !strncasecmp(last, "HTTP/0", 6))
    simple	= 1;

  for (i=0; i<sizeof meth/sizeof *meth; i++)
    if (!strcasecmp(s->s, meth[i].meth))
      {
	meth[i].fn(url, simple, meth[i].meth);
	return 0;
      }
  err(0, "unknown method", "%s", s->s);
  return 1;
}
